/*
   Situare - A location system for Facebook
   Copyright (C) 2010  Ixonos Plc. Authors:

       Ville Tiensuu - ville.tiensuu@ixonos.com
       Kaj Wallin - kaj.wallin@ixonos.com
       Henri Lampela - henri.lampela@ixonos.com

   Situare is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.

   Situare is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Situare; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
   USA.
*/

#ifndef FACEBOOKAUTHENTICATION_H
#define FACEBOOKAUTHENTICATION_H

#include <QUrl>

class QNetworkReply;
class QSslError;
class QWebView;

class MainWindow;

/**
* @brief FacebookAuthentication class takes care of Facebook login process. It creates
         QWebView instance and tries to login with cookies using hidden browser.
         If failed, then visible login browser dialog is invoked. Class also does parse the
         accuired credentials.
*
* @author Ville Tiensuu
* @author Sami Rämö - sami.ramo (at) ixonos.com
*/
class FacebookAuthentication : public QObject
{
    Q_OBJECT

public:
    /**
    * @brief Constructor
    *
    * Initiates internal data members.
    *
    * @param mainWindow MainWindow instance
    * @param parent Instance of the parent
    */
    FacebookAuthentication(MainWindow *mainWindow, QObject *parent = 0);

/*******************************************************************************
 * MEMBER FUNCTIONS AND SLOTS
 ******************************************************************************/
public slots:
    /**
    * @brief Clears account information from settings
    *
    * @param clearUserInformation True if user information should be cleared
    */
    void clearAccountInformation(bool clearUserInformation = false);

    /**
      * @brief Is the user currently logged in
      *
      * @returns True if the user is logged in, otherwise false
      */
    bool isLoggedIn() const;

    /**
      * @brief Initiate login process
      *
      * Builds login browser and starts loading login URL.
      */
    void login();

    /**
      * @brief Log out
      *
      * @param clearUserInformation True if user information should be cleared
      */
    void logOut(bool clearUserInformation = false);

private:
    /**
      * @brief Destroy login dialog and browser
      */
    void destroyLogin();

    /**
      * @brief Parses the session information from the URL
      *
      * @param url URL
      * @returns Parsed session, or empty string if parsing failed.
      */
    QString parseSession(const QUrl &url);

private slots:
    /**
      * @brief Cleanup after browser is destructed
      *
      * Clears the pointer to the browser and disables the progress indicator.
      */
    void browserDestroyed();

    /**
      * @brief Handler for login page loading errors
      *
      * @param reply Network reply
      */
    void networkReplyHandler(QNetworkReply *reply);

    /**
      * @brief Handler for SSL errors, ignores the error
      */
    void sslErrors(QNetworkReply *reply, const QList<QSslError> &errors);

    /**
      * @brief Handler for browser URL changes
      *
      * Does check the new URL and based on that invokes the login dialog with visible browser view
      * or parses the session from the new URL.
      *
      * @param url New URL
      */
    void urlChanged(const QUrl &url);

/*******************************************************************************
 * SIGNALS
 ******************************************************************************/
signals:
    /**
    * @brief Signals error
    *
    * @param context error context
    * @param error error code
    */
    void error(const int context, const int error);

    /**
      * @brief Emitted when logged in successfully
      *
      * All login related actions should be connected to this signal.
      *
      * @param session Session data
      * @param wallPostPermission Has the user granted rights for posting to wall
      */
    void loggedIn(const QString session, bool wallPostPermission);

    /**
      * @brief Emitted when logged out
      *
      * All logout related actions should be connected to this signal.
      */
    void loggedOut();

/*******************************************************************************
 * DATA MEMBERS
 ******************************************************************************/
private:
    bool m_loggedIn;                ///< Is the user currently logged in
    QWebView *m_browser;            ///< Login browser
    MainWindow *m_mainWindow;       ///< MainWindow
};

#endif // FACEBOOKAUTHENTICATION_H
