/*
   Situare - A location system for Facebook
   Copyright (C) 2010  Ixonos Plc. Authors:

      Henri Lampela - henri.lampela@ixonos.com
      Sami Rämö - sami.ramo@ixonos.com

   Situare is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.

   Situare is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Situare; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
   USA.
*/


#ifndef SITUARESERVICE_H
#define SITUARESERVICE_H

#include <QObject>

#include "../user/user.h"
#include "imagefetcher.h"

class NetworkAccessManager;
class QNetworkReply;
class QNetworkRequest;
class GeoCoordinate;
class QUrl;

/**
* @brief SituareService class for communicating with Situare server
*
* @author Henri Lampela
* @author Sami Rämö - sami.ramo (at) ixonos.com
*/
class SituareService : public QObject
{
    Q_OBJECT

public:

    /**
    * @brief Default constructor
    *
    * @param parent instance of parent
    */
    SituareService(QObject *parent = 0);

    /**
    * @brief Destructor
    *
    */
    ~SituareService();

/*******************************************************************************
 * MEMBER FUNCTIONS AND SLOTS
 ******************************************************************************/

    /**
    * @brief Retrieves location user and friends information from Situare server
    *
    */
    void fetchLocations();

    /**
    * @brief Translates coordinates to street address via Situare server
    *
    * @param coordinates coordinates to be translated
    */
    void reverseGeo(const GeoCoordinate &coordinates);

    /**
    * @brief Updates location to the Situare server
    *
    * @param coordinates current cordinates
    * @param status message
    * @param publish publish location on Facebook wall
    */
    void updateLocation(const GeoCoordinate &coordinates, const QString &status, const bool &publish);

public slots:

    /**
    * @brief Public slot, to clear user data
    *
    */
    void clearUserData();

    /**
    * @brief Public slot, which indicates when http request has been completed
    *
    * @param reply storage for http reply
    */
    void requestFinished(QNetworkReply *reply);

    /**
      * @brief Update session data
      *
      * @param session New session data
      */
    void updateSession(const QString &session);

private:

    /**
    * @brief Requests ImageFetcher if user/friend has a profile image
    *        uses members: m_user and m_friendsList
    *
    * @param imageUrlList list of image urls
    */
    void addProfileImages(const QList<QUrl> &imageUrlList);

    /**
      * @brief Append access_token and other session data to the reques url
      *
      * @param[in,out] requestUrl Request URL with other request parameters and ending to &
      */
    void appendAccessToken(QString &requestUrl);

    /**
      * @brief Convert coordinate value in degrees (double) to string with enough precision
      *
      * @param degrees Coordinate value in degrees
      * @returns Coordinate value as string
      */
    QString degreesToString(double degrees);

    /**
    * @brief Parses user and friend data from JSON string
    *
    * @param jsonReply JSON string
    */
    void parseUserData(const QByteArray &jsonReply);

    /**
      * @brief Build and send request
      *
      * Appends script pathname and parameters to the server base URL. Access token is appended and
      * the request sent if the access token is available, otherwise the request is queued.
      *
      * @param script Requested script pathname
      * @param parameters Hash containing parameter key/value pairs.
      */
    void buildRequest(const QString &script, const QHash<QString,QString> &parameters);

    /**
      * @brief Send request
      *
      * @param requestUrl Request URL containing also all parameters
      */
    void sendRequest(const QString &requestUrl);

private slots:
    /**
    * @brief Slot for received images
    *
    * @param url Image url
    * @param image Received image
    */
    void imageReceived(const QUrl &url, const QPixmap &image);

/*******************************************************************************
 * SIGNALS
 ******************************************************************************/

signals:

    /**
    * @brief Signals error
    *
    * @param context error context
    * @param error error code
    */
    void error(const int context, const int error);

    /**
    * @brief Signal for image fetching
    *
    * @param url Image url
    */
    void fetchImage(const QUrl &url);

    /**
    * @brief Signals when user's/friend's image is downloaded
    *
    * @param user Instance of user/friend
    */
    void imageReady(User *user);

    /**
    * @brief Signals when address data is retrieved
    *
    * @param address Street address
    */
    void reverseGeoReady(const QString &address);

    /**
    * @brief Signals when updateLocation request finished successfully
    *
    */
    void updateWasSuccessful();

    /**
    * @brief Signals when user data is retrieved
    *
    * @param user instance of user
    * @param friendList list of friends
    */
    void userDataChanged(User *user, QList<User *> &friendList);

/*******************************************************************************
 * DATA MEMBERS
 ******************************************************************************/

private:
    bool m_defaultImage;    ///< Indicates if some of the friends/user doesn't have a image

    QList<QNetworkReply *> m_currentRequests;       ///< List of current http requests
    QList<User *> m_friendsList;                    ///< List of friends(User)

    QString m_session;                          ///< Session data

    NetworkAccessManager *m_networkManager;     ///< Pointer to QNetworkAccessManager

    ImageFetcher *m_imageFetcher;               ///< Instance of the image fetcher
    User *m_user;                               ///< Pointer to User
};

#endif // SITUARESERVICE_H
